"""Useful clients for commonly accessed APIs/services"""

from .google_calendar import GoogleCalendarClient
from .google_drive import GoogleDriveClient
from .google_fit import GoogleFitClient
from .google_photos import GooglePhotosClient
from .monzo import MonzoClient
from .spotify import SpotifyClient
from .truelayer import TrueLayerClient
